typedef struct DwarfName DwarfName;
typedef struct DwarfOpName DwarfOpName;
typedef struct DwarfTagName DwarfTagName;
typedef struct DwarfAttrName DwarfAttrName;
typedef struct DwarfAttrValue DwarfAttrValue;
typedef struct DwarfCompileUnit DwarfCompileUnit;

struct DwarfName {
	int	tag;
	char	*name;
};

struct DwarfOpName {
	int	op;
	char	*name;
};

struct DwarfTagName {
	int	tag;
	char	*name;
};

struct DwarfAttrName {
	int	tag;
	char	*name;
};

struct DwarfAttrValue {
	Dwarf_Half	form;
	int		isvalid;
	int		issigned;

	Dwarf_Off 	off;
	Dwarf_Signed	sdata;
	Dwarf_Unsigned	udata;
	Dwarf_Addr	addr;
	Dwarf_Block	*block;
	char		*str;
};

struct DwarfCompileUnit {
	Dwarf_Half		addrsz;
	Dwarf_Half		version;
	Dwarf_Unsigned		hdrlen;
	Dwarf_Unsigned		offset;
	Dwarf_Unsigned		nextoff;
	Dwarf_Unsigned		abbrevoff;

	Dwarf_Signed		nsrcs;
	char			**srcs;

	Dwarf_Signed		nlines;
	Dwarf_Line		*lines;

	DwarfCompileUnit	*link;
};

enum {
	NoEndian	= 0,
	BigEndian,			/* If you use VAX, you lose */
	LittleEndian,

	BadStack	= -1,
	BadOp		= -2
};

extern	int		debug;
extern	int		verbose;
extern	int		dodups;
extern	int		endianness;
extern	char		*anonuniq;
extern	Dwarf_Error	dwarferror;

extern	DwarfName	*dwarfnametab;
extern	DwarfOpName	*dwarfopnametab;
extern	DwarfTagName	*dwarftagnametab;
extern	DwarfAttrName	*dwarfattrnametab;

DwarfName		*dwarfname(int);
DwarfOpName		*dwarfopname(int);
DwarfTagName		*dwarftagname(int);
DwarfAttrName		*dwarfattrname(int);

const char		*dwarfformname(int);

DwarfCompileUnit	*mkdwarfcompileunit(void);
void			dwarfcompileunitclear(Dwarf_Debug, DwarfCompileUnit*);

void			dwarfattrvalueprint(Dwarf_Debug, DwarfAttrValue*);
void			dwarfattrvalueclear(Dwarf_Debug, DwarfAttrValue*);
int			dwarfattrvalue(Dwarf_Debug, Dwarf_Die, Dwarf_Attribute, Dwarf_Half, DwarfAttrValue*, DwarfCompileUnit*);

int			dwarfevalconst(Dwarf_Debug, DwarfCompileUnit*, uchar*, int, uintptr*);

Dwarf_Signed		dwarfgetsignedleb128(uchar*, uchar*, int*);
Dwarf_Unsigned		dwarfgetunsignedleb128(uchar*, uchar*, int*);
uintptr			dwarfgetaddress(DwarfCompileUnit*, uchar*, uchar*, int*);

s8int			dwarfgetsigned8(uchar*, uchar*);
u8int			dwarfgetunsigned8(uchar*, uchar*);
s16int			dwarfgetsigned16(uchar*, uchar*);
u16int			dwarfgetunsigned16(uchar*, uchar*);
s32int			dwarfgetsigned32(uchar*, uchar*);
u32int			dwarfgetunsigned32(uchar*, uchar*);
s64int			dwarfgetsigned64(uchar*, uchar*);
u64int			dwarfgetunsigned64(uchar*, uchar*);

Dwarf_Signed		dwarfdecodesleb128(Dwarf_Small*, ulong*);
Dwarf_Unsigned		dwarfdecodeuleb128(Dwarf_Small*, ulong*);

int			dwarfinfowalk(Dwarf_Debug);
