typedef struct Decl Decl;
typedef struct Type Type;
typedef struct DwarfRef DwarfRef;
typedef struct DwarfVar DwarfVar;
typedef struct DwarfType DwarfType;
typedef struct DwarfStack DwarfStack;

typedef struct DwarfVarList DwarfVarList;
typedef struct DwarfTypeList DwarfTypeList;

/*
 * Id:			ref_addr
 * Unit+Offset: 	compile unit + offset (refN)
 */
struct DwarfType {
	ushort			tag;		/* DWARF tag */
	u64int			addr;		/* ID for lookup */
	u32int			offset;		/* Off in unit/CU */
	Type			*type;
	char			*file;
	ulong			line;
	TAILQ_ENTRY(DwarfType)	link;
	TAILQ_ENTRY(DwarfType)	hashlocal;	/* Chain for locals */
	TAILQ_ENTRY(DwarfType)	hashglobal;	/* Chain for globals */
	DwarfCompileUnit	*compile;
};

struct DwarfVar {
	u64int			addr;
	Decl			*decl;
	TAILQ_ENTRY(DwarfVar)	link;
};

struct DwarfRef {
	int			resolved;
	u64int			addr;
	u32int			offset;
	DwarfCompileUnit	*compile;
};

struct DwarfStack {
	int		depth;
	int		nalloc;
	DwarfType	**stk;
};

struct Type {
	int			type;
	u32int			flags;
	int			size;
	int			nalloc;		/* Allocated size of dim */
	int			narr;		/* Used size of dim */
	int			*dim;		/* Array dimensions */
	char			*tag;
        int                     encoding;
	Type			*sub;		/* Incl. func. return type */
	STAILQ_HEAD(, Decl)	members;	/* SUE members, formals */
	DwarfRef		dwarf;
};

struct Decl {
	u32int			offset;		/* Offset within struct/union */
	u16int			bitsize;	/* Width of bit field */
	u16int			bytesize;	/* Width of container */
	u16int			msboffset;	/* Bits to left of msb */
	u16int			bitoffset;	/* Bit offset less offset */
	char			*name;
	Type			*type;
	STAILQ_ENTRY(Decl)	link;
	DwarfAttrValue		value;		/* For enumerators */
	DwarfRef		dwarf;	
};

/* Make sure to update typetab[] if you change this */
enum {
	Tplastic	= 0,
	Tbase,			/* Dwarf provides base names */
	Tstruct,
	Tunion,
	Tenum,
	Tarr,
	Tptr,
	Ttypedef,

	Tfun,
	Tvararg,		/* Should have a single instance like void */

	Tref,
	Tconst,
	Tvolatile,
	Trestrict
};

enum {
	Fmark	= 1<<0,
	Fdecl	= 1<<1,		/* Type represents decl., not def. */
	Finline	= 1<<2
};

TAILQ_HEAD(DwarfVarList, DwarfVar);
TAILQ_HEAD(DwarfTypeList, DwarfType);

extern	char*		typetab[];
extern	DwarfType	*dwarfvoid;
extern	DwarfVarList	dwarfvars;
extern	DwarfTypeList	dwarftypes;

char			*anontype(void);
char			*anondecl(void);

DwarfStack		*mkdwarfstack(void);
void			dwarffreestack(DwarfStack*);
void			dwarfpush(DwarfStack*);
void			dwarfpop(DwarfStack*);
int			dwarfdepth(DwarfStack*);
void			dwarfsettop(DwarfStack*, DwarfType*);
void			dwarfdumpstack(DwarfStack*);
DwarfType		*dwarffindsueparent(DwarfStack*);
DwarfType		*dwarffindparent(DwarfStack*, int);

void			dwarfaddarraydim(Type*, u64int, u64int);

void			dwarfprinttypes(void);
int			dwarfresolvetypes(void);

void			dwarfprintsyms(void);
int			dwarfresolvesyms(void);

void			dwarftypeinit(void);
DwarfType		*mkdwarftype(DwarfCompileUnit*,ushort, u64int, u32int);
void			dwarftypeinsert(DwarfType*);
DwarfType		*dwarftypefindglobal(u64int);
DwarfType		*dwarftypefindlocal(u64int, u32int);

void			dwarfvarinit(void);
DwarfVar		*mkdwarfvar(Decl*, u64int);

DwarfType		*decl2dwarftype(Decl*);
DwarfType		*type2dwarftype(Type*);
int			dwarfresolvedecl(Decl*);

Type			*mktype(int);
Decl			*mkdecl(void);

void			printtype(Type*);
void			printsym(DwarfVar*);
